from dataclasses import dataclass
from urllib.parse import urlparse, urlunparse

from typing_extensions import Self


@dataclass
class URIComponents:
    scheme: str = ""
    netloc: str = ""
    path: str = ""
    params: str = ""
    query: str = ""
    fragment: str = ""

    def __str__(self):
        return urlunparse((self.scheme, self.netloc, self.path, self.params, self.query, self.fragment))

    def __next__(self):
        raise NotImplementedError("This is here for pandas...")

    @classmethod
    def from_str(cls, s: str) -> Self:
        return cls(*urlparse(s))



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, Hashable, List

import pandas as pd

from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType
from bridge.utils.constants import ELEMENT_COLS, INDICES
from bridge.utils.helper import Displayable

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE
    from bridge.primitives.sample.transform.sample_transform import SampleTransform


class Sample(Displayable):
    keys = ELEMENT_COLS.list()

    def __init__(
        self,
        elements: List[Element] | Dict[ElementType, List[Element]],
        display_engine: DisplayEngine | None = None,
    ):
        if isinstance(elements, dict):
            self._elements = elements
        else:
            self._elements = self._convert_elements_list_to_dict(elements)

        self._assert_valid_elements(self._elements)

        self._display_engine = display_engine

    @property
    def id(self) -> Hashable:
        for e_list in self._elements.values():
            if len(e_list) > 0:
                return e_list[0].sample_id

    @property
    def elements(self) -> Dict[ElementType, List[Element]]:
        return self._elements

    @property
    def data(self) -> Dict[ElementType, List[ELEMENT_DATA_TYPE]]:
        data_dict = defaultdict(list)
        for etype, elist in self._elements.items():
            data_dict[etype].extend([e.data for e in elist])
        return dict(data_dict)

    def show(self, **kwargs: Any):
        return self._display_engine.show_sample(self, **kwargs)

    def transform(
        self,
        transform: SampleTransform,
        cache_mechanisms: Dict[ElementType, CacheMechanism] | None = None,
        display_engine: DisplayEngine | None = None,
    ) -> "Sample":
        cache_mechanisms = self._get_cache_mechanisms_for_transform(self, cache_mechanisms)
        if display_engine is None:
            return transform(self, cache_mechanisms, self._display_engine)
        else:
            return transform(self, cache_mechanisms, display_engine)

    @classmethod
    def from_pd_dataframe(
        cls,
        elements_df: pd.DataFrame,
        display_engine: DisplayEngine | None,
        cache_mechanisms: Dict[ElementType, CacheMechanism | None],
    ):
        def fast_to_dict_records(df):
            data = df.values.tolist()
            columns = df.columns.tolist()
            index = df.index.names
            index_data = df.index.values.tolist()
            idx_records = [dict(zip(index, data)) for data in index_data]
            records = [dict(zip(columns, datum)) for datum in data]
            [rec.update(idx_rec) for rec, idx_rec in zip(records, idx_records)]
            return records

        elements = []
        for element_row in fast_to_dict_records(elements_df):
            etype = element_row[ELEMENT_COLS.ETYPE]

            element_type = ElementType(etype)
            elements.append(
                Element.from_dict(
                    element_row,
                    display_engine=display_engine,
                    cache_mechanism=cache_mechanisms.get(element_type),
                )
            )
        return cls(elements=elements, display_engine=display_engine)

    def to_pd_dataframe(self):
        records = []
        for e_list in self.elements.values():
            for element in e_list:
                records.append(element.to_pd_series())
        return pd.DataFrame(records).set_index(INDICES)

    def __len__(self) -> int:
        return sum(map(len, self._elements.values()))

    @staticmethod
    def _assert_valid_elements(elements: Dict[ElementType, List[Element]]):
        sample_ids_from_elements = set([e.sample_id for e_list in elements.values() for e in e_list])
        assert len(sample_ids_from_elements) == 1, (
            f"All elements must contain a single sample id,"
            f" got {len(sample_ids_from_elements)}: {sample_ids_from_elements}"
        )

    @staticmethod
    def _convert_elements_list_to_dict(elements: List[Element]) -> Dict[ElementType, List[Element]]:
        elements_by_type: Dict[ElementType, List[Element]] = defaultdict(list)
        for element in elements:
            elements_by_type[element.etype].append(element)
        d = dict(elements_by_type)
        return d

    @staticmethod
    def _get_cache_mechanisms_for_transform(sample: Sample, cache_mechanisms: Dict[ElementType, CacheMechanism] | None):
        if cache_mechanisms is None:
            cache_mechanisms = {}

        sample_etypes = sample.elements.keys()
        if set(sample_etypes) == set(cache_mechanisms.keys()):
            return cache_mechanisms

        default_cache_mechanisms = {etype: CacheMechanism() for etype in sample_etypes}

        missing_keys = set(default_cache_mechanisms.keys()) - set(cache_mechanisms.keys())
        if len(missing_keys) > 0:
            pass
            # warnings.warn(
            #     f"The following keys don't have cache mechanisms and are set to `memory` by default: {missing_keys}"
            # )
        default_cache_mechanisms.update(cache_mechanisms)
        return default_cache_mechanisms



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Hashable

import pandas as pd

from bridge.primitives.element.data.load.load_mechanism import LoadMechanism
from bridge.primitives.utils import validate_metadata
from bridge.utils.constants import ELEMENT_COLS
from bridge.utils.helper import Displayable

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
    from bridge.primitives.element.data.category import DataCategory
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE
    from bridge.primitives.element.element_type import ElementType


class Element(Displayable):
    keys = ELEMENT_COLS.list()

    def __init__(
        self,
        element_id: Hashable,
        etype: ElementType,
        load_mechanism: LoadMechanism,
        sample_id: Hashable,
        display_engine: DisplayEngine | None = None,
        cache_mechanism: CacheMechanism | None = None,
        metadata: Dict[str, Any] | None = None,
    ):
        validate_metadata(self.keys, metadata)
        self._element_id = element_id
        self._etype = etype
        self._sample_id = sample_id
        self._load_mechanism = load_mechanism
        self._display_engine = display_engine
        self._cache_mechanism = cache_mechanism
        self._metadata = metadata or {}

    @property
    def id(self) -> Hashable:
        return self._element_id

    @property
    def data(self) -> ELEMENT_DATA_TYPE:
        return self._data_impl()

    def _data_impl(self):
        data = self._load_mechanism.load_data()
        if self._cache_mechanism:
            new_load_mechanism = self._cache_mechanism.store(self, data, should_update_elements=True)
            self._load_mechanism = new_load_mechanism
            return data
        return data

    @property
    def etype(self) -> ElementType:
        return self._etype

    @property
    def category(self) -> DataCategory:
        return self._load_mechanism.category

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    def sample_id(self) -> Hashable:
        return self._sample_id

    def __str__(self) -> str:
        return str(self.to_dict())

    def to_dict(self) -> Dict[str, Any]:
        return {
            ELEMENT_COLS.ID: self.id,
            ELEMENT_COLS.ETYPE: self.etype,
            ELEMENT_COLS.SAMPLE_ID: self.sample_id,
            **self._load_mechanism.to_dict(),
            **self.metadata,
        }

    @classmethod
    def from_dict(cls, dic: Dict[str, Any], **kwargs):
        assert set(dic.keys()).issuperset(set(cls.keys)), f"Missing keys: {set(cls.keys) - set(dic.keys())}"
        element_id = dic[ELEMENT_COLS.ID]
        sample_id = dic[ELEMENT_COLS.SAMPLE_ID]
        etype = dic[ELEMENT_COLS.ETYPE]
        load_mechanism = LoadMechanism.from_dict({k: v for k, v in dic.items() if k in LoadMechanism.keys})
        metadata = {k: v for k, v in dic.items() if k not in cls.keys}
        return cls(
            element_id=element_id,
            etype=etype,
            load_mechanism=load_mechanism,
            sample_id=sample_id,
            display_engine=kwargs.get("display_engine"),
            cache_mechanism=kwargs.get("cache_mechanism"),
            metadata=metadata,
        )

    def to_pd_series(self) -> pd.Series:
        return pd.Series(self.to_dict())

    @classmethod
    def from_pd_series(
        cls,
        element_series: pd.Series,
        display_engine: DisplayEngine | None = None,
        cache_mechanism: CacheMechanism | None = None,
    ):
        return cls.from_dict(
            {**element_series.to_dict()},
            display_engine=display_engine,
            cache_mechanism=cache_mechanism,
        )

    def show(self, **kwargs):
        return self._display_engine.show_element(self, **kwargs)



from typing import TYPE_CHECKING, Any, Dict

import pandas as pd
import panel as pn

from bridge.primitives.element.element_type import ElementType
from bridge.primitives.sample.singular_sample import SingularSample

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.dataset import SingularDataset
    from bridge.primitives.element.element import Element


class TextClassification(DisplayEngine[SingularDataset, SingularSample]):
    def show_element(self, element: Element, element_plot_kwargs: Dict[str, Any] | None = None):
        if element.etype == ElementType.class_label:
            return pn.pane.Markdown(element.to_pd_series().to_frame().T.to_markdown())
        elif element.etype == ElementType.text:
            return pn.pane.Markdown(element.data)
        else:
            raise NotImplementedError()

    def show_sample(
        self,
        sample: SingularSample,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
    ):
        annotations_md = pd.DataFrame(
            [ann.to_pd_series() for ann in sample.annotations[ElementType.class_label]]
        ).to_markdown()
        text_display = pn.pane.Markdown(sample.data)
        return pn.Column("# Sample Text:", text_display, "# Annotations Data:", annotations_md)

    def show_dataset(
        self,
        dataset: SingularDataset,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
        dataset_plot_kwargs: Dict[str, Any] | None = None,
    ):
        sample_ids = dataset.sample_ids
        sample_ids_wig = pn.widgets.DiscreteSlider(name="Sample ID", options=sample_ids, value=sample_ids[0])

        @pn.depends(sample_ids_wig.param.value)
        def plot_sample_by_widget(sample_id):
            return self.show_sample(dataset.get(sample_id), element_plot_kwargs, sample_plot_kwargs)

        return pn.Column(sample_ids_wig, plot_sample_by_widget)


# class Panel(DisplayEngine):
#     def show_element(self, element: Element, element_plot_kwargs: Dict[str, Any] | None = None):
#         if element.etype == ElementType.class_label:
#             return self._show_class_label(element, element_plot_kwargs)
#         elif element.etype == ElementType.text:
#             return self._show_text(element, element_plot_kwargs)
#         else:
#             raise NotImplementedError()
#
#     def show_sample(
#         self,
#         sample: Sample,
#         element_plot_kwargs: Dict[str, Any] | None = None,
#         sample_plot_kwargs: Dict[str, Any] | None = None,
#     ):
#         annotations_md = pd.DataFrame(
#             [ann.to_pd_series() for ann in sample.elements[ElementType.class_label]]
#         ).to_markdown()
#         text_display = self.show_element(sample.elements[ElementType.text][0])
#         return pn.Column("# Sample Text:", text_display, "# Annotations Data:", annotations_md)
#
#     def show_dataset(
#         self,
#         dataset: Dataset,
#         element_plot_kwargs: Dict[str, Any] | None = None,
#         sample_plot_kwargs: Dict[str, Any] | None = None,
#         dataset_plot_kwargs: Dict[str, Any] | None = None,
#     ):
#         sample_ids = dataset.sample_ids
#         sample_ids_wig = pn.widgets.DiscreteSlider(name="Sample ID", options=sample_ids, value=sample_ids[0])
#
#         @pn.depends(sample_ids_wig.param.value)
#         def plot_sample_by_widget(sample_id):
#             return self.show_sample(dataset.get(sample_id), element_plot_kwargs, sample_plot_kwargs)
#
#         return pn.Column(sample_ids_wig, plot_sample_by_widget)
#
#     @staticmethod
#     def _show_class_label(element: Element, element_plot_kwargs: Dict[str, Any] | None):
#         return pn.pane.Markdown(element.to_pd_series().to_frame().T.to_markdown())
#
#     @staticmethod
#     def _show_text(element: Element, element_plot_kwargs: Dict[str, Any] | None):
#         return pn.pane.Markdown(element.data)



from bridge.utils.download_and_extract_archive import download_and_extract_archive
from bridge.utils.helper import Dictable, StrEnum
from bridge.utils.optional_import import optional_dependencies
from bridge.utils.pmap import pmap

__all__ = ["Dictable", "download_and_extract_archive", "StrEnum", "optional_dependencies", "pmap"]



import numpy as np
import pytest

from bridge.display import DisplayEngine
from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.data.load.load_mechanism import LoadMechanism
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType
from bridge.utils.constants import ELEMENT_COLS
from bridge.utils.data_objects import BoundingBox, ClassLabel, Keypoint


@pytest.fixture(
    params=[
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: DataCategory.obj,
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: BoundingBox(np.array([0, 0, 1, 1]), class_label=ClassLabel(0)),
        },
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: DataCategory.obj,
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: ClassLabel(class_idx=0, class_name="some_class"),
        },
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: DataCategory.image,
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: "http://example.com/image.jpg",
        },
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: DataCategory.image,
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: "dummy_path.jpg",
        },
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: DataCategory.obj,
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: Keypoint(np.array([0, 0])),
        },
    ],
    ids=[
        f"{DataCategory.obj}_memory_bbox",
        f"{DataCategory.obj}_memory_class_label",
        f"{DataCategory.image}_http",
        f"{DataCategory.image}_file",
        f"{DataCategory.obj}_memory_keypoint",
    ],
)
def load_mechanism_dict(request):
    return request.param


@pytest.fixture(
    params=[
        {
            ELEMENT_COLS.LOAD_MECHANISM.CATEGORY: "dummy",
            "lol": "dummy",
        },
        {
            "lol": "dummy",
            ELEMENT_COLS.LOAD_MECHANISM.URL_OR_DATA: "dummy",
        },
    ],
    ids=["no_url_or_data", "no_category"],
)
def bad_load_mechanism_dict(request):
    return request.param


@pytest.fixture
def element_dict(load_mechanism_dict):
    return {
        ELEMENT_COLS.ID: "123",
        ELEMENT_COLS.ETYPE: ElementType.image,
        ELEMENT_COLS.SAMPLE_ID: 0,
        **load_mechanism_dict,
    }


def test_load_mechanism_from_dict(load_mechanism_dict):
    lm = LoadMechanism.from_dict(load_mechanism_dict)
    assert load_mechanism_dict == lm.to_dict()


def test_bad_load_mechanism_dict(bad_load_mechanism_dict):
    with pytest.raises(AssertionError):
        LoadMechanism.from_dict(bad_load_mechanism_dict)


def test_element_from_dict(mocker, element_dict):
    de_mock = mocker.Mock(spec=DisplayEngine)
    cm_mock = mocker.Mock(spec=CacheMechanism)
    element = Element.from_dict(element_dict, display_engine=de_mock, cache_mechanism=cm_mock)
    dic = element.to_dict()
    assert dic == element_dict



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Dict

import numpy as np
from tqdm import tqdm

from bridge.display.basic import SimplePrints
from bridge.primitives.dataset import SingularDataset
from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.data.load.load_mechanism import LoadMechanism
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType
from bridge.primitives.sample.singular_sample import SingularSample
from bridge.providers.dataset_provider import DatasetProvider
from bridge.utils import download_and_extract_archive, optional_dependencies
from bridge.utils.data_objects import BoundingBox, ClassLabel

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism


class ImageFolder(DatasetProvider[SingularDataset, SingularSample]):
    def __init__(self, root: str | os.PathLike):
        self._root = root

    def build_dataset(
        self, display_engine: DisplayEngine = None, cache_mechanisms: Dict[ElementType, CacheMechanism] = None
    ):
        images = []
        classes = []
        for i, class_dir in enumerate(sorted(Path(self._root).iterdir())):
            for img_file in class_dir.iterdir():
                sample_id = len(images)
                img_element = Element(
                    element_id=f"image_{sample_id}",
                    sample_id=sample_id,
                    etype=ElementType.image,
                    load_mechanism=LoadMechanism.from_url_string(str(img_file), category=DataCategory.image),
                    metadata={"filename": img_file.name},
                )
                class_element = Element(
                    element_id=f"class_{i}",
                    sample_id=sample_id,
                    etype=ElementType.class_label,
                    load_mechanism=LoadMechanism(ClassLabel(i, class_dir.name), category=DataCategory.obj),
                    metadata={"filename": img_file.name},
                )
                images.append(img_element)
                classes.append(class_element)
        return SingularDataset.from_lists(
            images, classes, display_engine=display_engine, cache_mechanisms=cache_mechanisms
        )


class Coco2017Detection(DatasetProvider[SingularDataset, SingularSample]):
    images_download_links = {
        "train": "http://images.cocodataset.org/zips/train2017.zip",
        "val": "http://images.cocodataset.org/zips/val2017.zip",
    }

    annotations_download_links = {
        "train": "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
        "val": "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
    }
    images_source = {"download", "local", "stream"}

    def __init__(self, root: str | os.PathLike, split: str = "train", img_source: str = "stream"):
        from pycocotools.coco import COCO

        assert split in self.images_download_links.keys(), f"Split must be one of: {self.images_download_links.keys()}"

        root = Path(root).expanduser()
        self._images_dir = root / f"{split}2017"
        self._img_source = img_source
        self._ann_file = root / "annotations" / f"instances_{split}2017.json"

        if self._ann_file.exists():
            print(f"Annotations file {self._ann_file} already exists, skipping download.")
        else:
            print("Downloading annotations...")
            download_and_extract_archive(self.annotations_download_links[split], str(root))

        if self._img_source == "download":
            root.mkdir(parents=True, exist_ok=True)
            if self._images_dir.exists():
                print(f"Images directory {self._images_dir} already exists, skipping download.")
            else:
                print("Downloading images...")
                download_and_extract_archive(self.images_download_links[split], str(root))
        self._coco = COCO(self._ann_file)

    def build_dataset(
        self,
        display_engine: DisplayEngine = SimplePrints(),
        cache_mechanisms: Dict[ElementType, CacheMechanism] = None,
    ):
        img_id_list = list(sorted(self._coco.imgs.keys()))
        images = []
        bboxes = []
        for img_id in tqdm(img_id_list):
            coco_img = self._coco.loadImgs(img_id)[0]
            img_file = self._images_dir / coco_img["file_name"]

            if self._img_source == "stream":
                url = coco_img["coco_url"]  # noqa
            else:
                url = str(img_file)
            load_mechanism = LoadMechanism.from_url_string(url, category=DataCategory.image)  # noqa
            img_element = Element(
                element_id=f"{img_id}_img",
                sample_id=img_id,
                etype=ElementType.image,
                load_mechanism=load_mechanism,
                metadata={k: v for k, v in coco_img.items() if k not in Element.keys and k != "id"},
            )
            images.append(img_element)
            coco_annotations = self._coco.loadAnns(self._coco.getAnnIds(img_id))
            for coco_ann_dict in coco_annotations:
                category_id = coco_ann_dict["category_id"]
                bbox_data = BoundingBox(coords=(np.array(coco_ann_dict["bbox"])), class_label=(ClassLabel(category_id)))
                load_mechanism = LoadMechanism(bbox_data, category=DataCategory.obj)
                bbox_element = Element(
                    element_id=f"{img_id}_{coco_ann_dict['id']}",
                    sample_id=img_id,
                    etype=ElementType.bbox,
                    load_mechanism=load_mechanism,
                    metadata={
                        "category_id": category_id,
                        "area": coco_ann_dict["area"],
                        "iscrowd": coco_ann_dict["iscrowd"],
                    },
                )
                bboxes.append(bbox_element)
        return SingularDataset.from_lists(
            images, bboxes, display_engine=display_engine, cache_mechanisms=cache_mechanisms
        )


class TorchvisionCIFAR10(DatasetProvider[SingularDataset, SingularSample]):
    def __init__(self, root: str | os.PathLike, train: bool = True, download: bool = False):
        with optional_dependencies("raise"):
            from torchvision.datasets import CIFAR10

        self._ds = CIFAR10(root=root, train=train, download=download)

    def build_dataset(
        self,
        display_engine: DisplayEngine = SimplePrints(),
        cache_mechanisms: Dict[ElementType, CacheMechanism | None] | None = None,
    ):
        sample_list = []
        annotation_list = []
        for i, (img, target) in enumerate(zip(self._ds.data, self._ds.targets)):
            img_element = Element(
                element_id=i,
                etype=ElementType.image,
                sample_id=i,
                load_mechanism=LoadMechanism(url_or_data=img, category=DataCategory.image),
            )
            label_element = Element(
                element_id=f"label_{i}",
                etype=ElementType.class_label,
                sample_id=i,
                load_mechanism=LoadMechanism(
                    url_or_data=ClassLabel(class_idx=target, class_name=self._ds.classes[target]),
                    category=DataCategory.obj,
                ),
            )
            sample_list.append(img_element)
            annotation_list.append(label_element)

        return SingularDataset.from_lists(
            sample_list, annotation_list, display_engine=display_engine, cache_mechanisms=cache_mechanisms
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Generic

from bridge.primitives.types import D, S

if TYPE_CHECKING:
    from bridge.primitives.element.element import Element


class DisplayEngine(ABC, Generic[D, S]):
    @abstractmethod
    def show_element(
        self,
        element: Element,
        element_plot_kwargs: Dict[str, Any] | None = None,
    ):
        pass

    @abstractmethod
    def show_sample(
        self,
        sample: S,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
    ):
        pass

    @abstractmethod
    def show_dataset(
        self,
        dataset: D,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
        dataset_plot_kwargs: Dict[str, Any] | None = None,
    ):
        pass



import pytest

from bridge.display import DisplayEngine
from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
from bridge.primitives.element.data.load.load_mechanism import LoadMechanism
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType


@pytest.fixture
def dummy_element_id():
    return 0


@pytest.fixture
def dummy_sample_id():
    return "0"


@pytest.fixture
def dummy_etype():
    return ElementType.segmentation


@pytest.fixture
def load_mechanism_mock(mocker):
    mock = mocker.Mock(spec=LoadMechanism)
    mock.category.return_value = "some_category"
    mock.load_data.return_value = b"some_data"
    return mock


@pytest.fixture
def display_engine_mock(mocker):
    return mocker.Mock(spec=DisplayEngine)


@pytest.fixture
def cache_mechanism_mock(mocker):
    return mocker.Mock(spec=CacheMechanism)


@pytest.fixture
def dummy_metadata():
    return {"lol": "kek", "foo": 5}


@pytest.fixture
def dummy_element(
    dummy_element_id,
    dummy_sample_id,
    dummy_etype,
    load_mechanism_mock,
    display_engine_mock,
    cache_mechanism_mock,
    dummy_metadata,
):
    return Element(
        element_id=dummy_element_id,
        etype=dummy_etype,
        load_mechanism=load_mechanism_mock,
        sample_id=dummy_sample_id,
        display_engine=display_engine_mock,
        cache_mechanism=cache_mechanism_mock,
        metadata=dummy_metadata,
    )


def test_validate_metadata(load_mechanism_mock, display_engine_mock, cache_mechanism_mock):
    with pytest.raises(AssertionError):
        Element(
            element_id=0,
            sample_id=0,
            etype=ElementType.segmentation,
            load_mechanism=load_mechanism_mock,
            display_engine=display_engine_mock,
            cache_mechanism=cache_mechanism_mock,
            metadata={"source": "lol", "data": "dsadsa"},
        )


def test_element_static_properties(
    dummy_element, dummy_element_id, dummy_sample_id, dummy_etype, load_mechanism_mock, dummy_metadata
):
    assert dummy_element.id == dummy_element_id
    assert dummy_element.sample_id == dummy_sample_id
    assert dummy_element.etype == dummy_etype
    assert dummy_element.metadata == dummy_metadata
    assert dummy_element.category == load_mechanism_mock.category


def test_element_data(dummy_element, load_mechanism_mock, cache_mechanism_mock):
    assert dummy_element.data == load_mechanism_mock.load_data()
    assert cache_mechanism_mock.store.called



from __future__ import annotations

import functools
from types import GeneratorType
from typing import TYPE_CHECKING, Any, Callable, Dict, Hashable, Iterable, Iterator, List, Sequence

import pandas as pd
from tqdm.contrib import tmap
from typing_extensions import Self

from bridge.primitives.dataset.sample_api import SampleAPI
from bridge.primitives.dataset.table_api import TableAPI
from bridge.primitives.element.element_type import ElementType
from bridge.primitives.sample import Sample
from bridge.utils.constants import ELEMENT_COLS, INDICES
from bridge.utils.helper import Displayable

if TYPE_CHECKING:
    from bridge.display.display_engine import DisplayEngine
    from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
    from bridge.primitives.element.element import Element
    from bridge.primitives.sample.transform import SampleTransform


class Dataset(TableAPI, SampleAPI, Displayable):
    def __init__(
        self,
        elements: pd.DataFrame,
        display_engine: DisplayEngine = None,
        cache_mechanisms: Dict[ElementType, CacheMechanism | None] | None = None,
    ):
        self._elements = elements
        self._display_engine = display_engine
        self._cache_mechanisms = cache_mechanisms or {}
        self._connect_caches()

    @property
    def elements(self) -> pd.DataFrame:
        return self._elements.copy()

    @property
    def sample_ids(self) -> List[Hashable]:
        return self._elements.index.get_level_values(ELEMENT_COLS.SAMPLE_ID).drop_duplicates().to_list()

    def select(self, selector: Callable):
        selected = selector(self.elements)
        elements = self.elements.loc[selected]
        return Dataset(elements, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms)

    def assign(self, **kwargs: Dict[str, Callable[[pd.DataFrame], Sequence]]) -> Self:
        new_elements = self._elements.assign(**kwargs)
        return Dataset(new_elements, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms)

    def sort(self, by: str, ascending: bool = True):
        new_elements = self._elements.sort_values(by=by, ascending=ascending)
        return Dataset(new_elements, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms)

    def merge(
        self,
        other: "Dataset",
        display_engine: DisplayEngine | None = None,
        cache_mechanisms: Dict[ElementType, CacheMechanism | None] | None = None,
    ) -> "Dataset":
        self_element_ids = self.elements.index.get_level_values(ELEMENT_COLS.ID)
        other_element_ids = other.elements.index.get_level_values(ELEMENT_COLS.ID)
        assert (
            len(self_element_ids.intersection(other_element_ids)) == 0
        ), "Cannot merge Datasets with duplicate element ids."
        elements = pd.concat([self.elements, other.elements])
        if display_engine is None:
            display_engine = self._display_engine
        if cache_mechanisms is None:
            cache_mechanisms = self._cache_mechanisms
        return Dataset(elements, display_engine, cache_mechanisms=cache_mechanisms)

    def iget(self, index: int) -> Sample:
        sample_id = self.sample_ids[index]
        return self.get(sample_id)

    def get(self, sample_id: Hashable) -> Sample:
        sample_df = self._elements.xs(sample_id, level=ELEMENT_COLS.SAMPLE_ID, drop_level=False)
        return Sample.from_pd_dataframe(
            sample_df, display_engine=self._display_engine, cache_mechanisms=self._cache_mechanisms
        )

    def transform_samples(
        self,
        transform: SampleTransform,
        map_fn=tmap,
        cache_mechanisms: Dict[ElementType, CacheMechanism] | None = None,
        display_engine: DisplayEngine | None = None,
    ) -> Self:
        fn = functools.partial(
            Sample.transform, transform=transform, cache_mechanisms=cache_mechanisms, display_engine=display_engine
        )
        samples = map_fn(fn, self)
        if isinstance(samples, GeneratorType):
            samples = list(samples)
        elements = [element for sample in samples for e_list in sample.elements.values() for element in e_list]
        return Dataset.from_elements(elements, display_engine=display_engine)

    def map_samples(self, function: Callable[[Sample], Any], map_fn=tmap):
        outputs = map_fn(function, self)
        if isinstance(outputs, GeneratorType):
            return list(outputs)
        return outputs

    def show(self, **kwargs):
        return self._display_engine.show_dataset(self, **kwargs)

    def __getitem__(self, item):
        return self.iget(item)

    def __iter__(self) -> Iterator[Sample]:
        for sample_id in self.sample_ids:
            yield self.get(sample_id)

    def __len__(self) -> int:
        sample_ids = self.sample_ids
        return len(sample_ids)

    def __repr__(self) -> str:
        lens_dict = {"n_samples": len(self)}
        for etype, group in self._elements.groupby(ELEMENT_COLS.ETYPE):
            lens_dict[f"n_{etype}"] = len(group)
        return "Dataset: " + str(lens_dict)

    @classmethod
    def from_elements(
        cls,
        elements: Iterable[Element],
        display_engine: DisplayEngine = None,
        cache_mechanisms: Dict[ElementType, CacheMechanism | None] | None = None,
    ) -> Self:
        element_records = [e.to_pd_series() for e in elements]
        elements_df = pd.DataFrame(element_records).set_index(INDICES)
        return cls(elements=elements_df, display_engine=display_engine, cache_mechanisms=cache_mechanisms)

    def _connect_caches(self):
        for cache in self._cache_mechanisms.values():
            if cache is not None:
                cache.set_elements_df(self._elements)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict

import numpy as np

from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.data.uri_components import URIComponents

if TYPE_CHECKING:
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE


class LoadCategory(ABC):
    @staticmethod
    @abstractmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        pass


class LoadImage(LoadCategory):
    @staticmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return np.array(url_or_data)  # assumes object is a PIL image or np.ndarray

        if url_or_data.scheme not in ["http", "https", "file", ""]:
            raise NotImplementedError()
        from skimage.io import imread

        return imread(str(url_or_data))


class LoadText(LoadCategory):
    @staticmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data  # assumes object is a string

        return open(str(url_or_data), "r").read()


class LoadTorch(LoadCategory):
    @staticmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data
        import torch

        return torch.load(str(url_or_data))


class LoadNumpy(LoadCategory):
    @staticmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data  # assume that is already np array
        return np.load(str(url_or_data))


class LoadObj(LoadCategory):
    @staticmethod
    def load(url_or_data: URIComponents | ELEMENT_DATA_TYPE) -> ELEMENT_DATA_TYPE:
        if not isinstance(url_or_data, URIComponents):
            return url_or_data  # assume that is already np array
        raise NotImplementedError()


class LoadingMethodExecutor:
    LOAD_DATA_METHODS: Dict[DataCategory, LoadCategory] = {
        DataCategory.image: LoadImage,
        DataCategory.text: LoadText,
        DataCategory.torch: LoadTorch,
        DataCategory.numpy: LoadNumpy,
        DataCategory.obj: LoadObj,
    }

    def load(self, url_or_data: URIComponents | ELEMENT_DATA_TYPE, category: DataCategory) -> ELEMENT_DATA_TYPE:
        return self.LOAD_DATA_METHODS[category].load(url_or_data)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from torch.utils.data import Dataset as PytorchDataset

if TYPE_CHECKING:
    from bridge.primitives.dataset import Dataset


class PytorchEngineDataset(PytorchDataset):
    def __init__(self, dataset: Dataset) -> None:
        self._dataset = dataset

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index) -> Any:
        elements = self._dataset.iget(index).elements
        outs = defaultdict(list)
        for etype, e_list in elements.items():
            for e in e_list:
                outs[str(etype)].append(e.data)
        return dict(outs)



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Dict, List, Union

import albumentations as A
import numpy as np
from PIL.Image import Image

from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType
from bridge.primitives.sample import Sample
from bridge.primitives.sample.transform.sample_transform import SampleTransform
from bridge.utils import optional_dependencies
from bridge.utils.data_objects import BoundingBox

if TYPE_CHECKING:
    from bridge.display import DisplayEngine
    from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism


class AlbumentationsCompose(SampleTransform):
    def __init__(
        self,
        albm_transforms: List[A.BasicTransform],
        bbox_format="pascal_voc",
        kp_format="xy",
    ) -> None:
        self._bbox_format = bbox_format
        self._kp_format = kp_format
        self._transforms = albm_transforms

    def __call__(
        self,
        sample: Sample,
        cache_mechanisms: Dict[ElementType, CacheMechanism],
        display_engine: DisplayEngine | None,
    ) -> Sample:
        elements = copy.deepcopy(sample.elements)
        albm_dict = self._elements_to_albm(elements)
        targets_dict = {k: k.split("_")[0] for k in albm_dict.keys()}
        compose = A.Compose(
            self._transforms,
            bbox_params=A.BboxParams(format=self._bbox_format),
            keypoint_params=A.KeypointParams(format=self._kp_format),
            additional_targets=targets_dict,
            is_check_shapes=False,
        )
        albm_dict = compose(**albm_dict)
        elements = self._albm_to_elements(elements, albm_dict, cache_mechanisms)
        sample = Sample(elements=elements, display_engine=display_engine)
        return sample

    def _elements_to_albm(self, elements: Dict[ElementType, List[Element]]):
        albm_dict = {}
        assert ElementType.image in elements, "Can't apply albumentations without an image element."
        for i, img_element in enumerate(elements[ElementType.image]):
            data = img_element.data
            albm_dict[f"image_{i}"] = data
        albm_dict.update({"image": albm_dict["image_0"], "bboxes": [], "keypoints": []})

        if ElementType.bbox in elements:
            for i, bbox_element in enumerate(elements[ElementType.bbox]):
                data: BoundingBox = bbox_element.data
                albm_dict[f"bboxes_{i}"] = [[*(data.coords.tolist()), data.class_label]]
        if ElementType.keypoint in elements:
            raise NotImplementedError("Didn't fully implement keypoints in albumentations yet.")
            # for i, keypoint in enumerate(elements[ElementType.keypoint]):
            #     keypoint: Keypoint
            #     albm_dict[f"keypoints_{i}"] = keypoint.coords
        return albm_dict

    def _albm_to_elements(
        self,
        elements: Dict[ElementType, List[Element]],
        albm_dict: Dict[str, Any],
        cache_mechanisms: Dict[ElementType, CacheMechanism],
    ):
        albm_to_elements = {"bboxes": ElementType.bbox, "keypoints": ElementType.keypoint, "image": ElementType.image}
        del albm_dict["image"]
        del albm_dict["bboxes"]
        del albm_dict["keypoints"]
        sorted_desc_albm_keys = sorted(albm_dict.keys(), key=lambda k: int(k.split("_")[1]), reverse=True)
        for albm_key in sorted_desc_albm_keys:
            albm_data = albm_dict[albm_key]
            albm_type, element_idx = albm_key.split("_")
            element_idx = int(element_idx)
            curr_element = elements[albm_to_elements[albm_type]][element_idx]
            if len(albm_data) == 0:  # deleted element
                del elements[albm_to_elements[albm_type]][element_idx]
            else:
                curr_element = self._update_element_with_transformed_data(albm_data, cache_mechanisms, curr_element)
                elements[curr_element.etype][element_idx] = curr_element

        return elements

    @staticmethod
    def _update_element_with_transformed_data(
        albm_data: Union[Image, List, np.ndarray],
        cache_mechanisms: Dict[ElementType, CacheMechanism],
        curr_element: Element,
    ):
        if curr_element.etype == ElementType.bbox:
            albm_data = np.array(albm_data[0])
            new_element_data = BoundingBox(albm_data[:4], class_label=albm_data[4])  # noqa
            new_category = DataCategory.obj
        elif curr_element.etype == ElementType.image:
            if isinstance(albm_data, np.ndarray):
                new_category = DataCategory.image
            else:
                with optional_dependencies(error="raise"):
                    import torch
                if isinstance(albm_data, torch.Tensor):
                    new_category = DataCategory.torch
                else:
                    raise NotImplementedError(f"invalid data type: {type(albm_data)}")
            new_element_data = albm_data
        else:
            raise NotImplementedError()

        provider = cache_mechanisms[curr_element.etype].store(
            curr_element, new_element_data, as_category=new_category, should_update_elements=False
        )
        curr_element = Element(
            element_id=curr_element.id,
            etype=curr_element.etype,
            load_mechanism=provider,
            sample_id=curr_element.sample_id,
            metadata=curr_element.metadata,
        )
        return curr_element



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Dict

from bridge.display.display_engine import DisplayEngine

if TYPE_CHECKING:
    from bridge.primitives.dataset import Dataset
    from bridge.primitives.element.element import Element
    from bridge.primitives.sample import Sample


class SimplePrints(DisplayEngine):
    def show_element(
        self,
        element: Element,
        element_plot_kwargs: Dict[str, Any] | None = None,
    ):
        print_dict = element.to_dict()
        print(json.dumps(print_dict, indent=4, default=str))

    def show_sample(
        self,
        sample: Sample,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
    ):
        element_dicts = {}
        for e_type, e_list in sample.elements.items():
            element_dicts[f"etype={e_type}"] = [e.to_dict() for e in e_list]

        print_dict = {f"Sample ID {sample.id}": {"Elements": element_dicts}}
        print(json.dumps(print_dict, indent=4, default=str))

    def show_dataset(
        self,
        dataset: Dataset,
        element_plot_kwargs: Dict[str, Any] | None = None,
        sample_plot_kwargs: Dict[str, Any] | None = None,
        dataset_plot_kwargs: Dict[str, Any] | None = None,
    ):
        for sample in dataset:
            self.show_sample(sample)
            print()



import pytest

from bridge.utils import optional_dependencies
from bridge.utils.optional_import import _generate_msg  # noqa


@pytest.fixture(autouse=True)
def assert_missing_module():
    with pytest.raises(ImportError):
        import missing_module  # noqa


def test_missing_module_raise():
    with pytest.raises(ImportError) as e_info:
        with optional_dependencies("raise"):
            import missing_module  # noqa
    assert e_info.value.msg == _generate_msg("missing_module", None)


def test_missing_module_raise_extra():
    with pytest.raises(ImportError) as e_info:
        with optional_dependencies("raise", ["dev"]):
            import missing_module  # noqa
    assert e_info.value.msg == _generate_msg("missing_module", ["dev"])


def test_missing_module_warn():
    with pytest.warns(UserWarning, match=_generate_msg("missing_module", None)):
        with optional_dependencies("warn"):
            import missing_module  # noqa


def test_missing_module_ignore(assert_missing_module):
    with optional_dependencies("ignore"):
        import missing_module  # noqa



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Dict

from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.data.load.load_mechanism import LoadMechanism

if TYPE_CHECKING:
    from bridge.primitives.element.data.uri_components import URIComponents
    from bridge.primitives.element.element_data_type import ELEMENT_DATA_TYPE


class CacheCategory(ABC):
    @staticmethod
    @abstractmethod
    def store(
        data: ELEMENT_DATA_TYPE,
        url: URIComponents | None,
        category: DataCategory,
    ) -> LoadMechanism: ...


class CacheImage(CacheCategory):
    @staticmethod
    def store(data: ELEMENT_DATA_TYPE, url: URIComponents | None, category: DataCategory) -> LoadMechanism:
        import PIL.Image
        from skimage.io import imsave

        if url is None:
            return LoadMechanism(PIL.Image.fromarray(data), category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError()

        path = Path(str(url)).expanduser()

        Path.mkdir(path.parent, parents=True, exist_ok=True)
        imsave(path, data)
        return LoadMechanism.from_url_string(str(url), category)


class CacheTorch(CacheCategory):
    @staticmethod
    def store(data: ELEMENT_DATA_TYPE, url: URIComponents | None, category: DataCategory) -> LoadMechanism:
        if url is None:
            return LoadMechanism(data, category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError()
        import torch

        path = Path(str(url))

        Path.mkdir(path.parent, parents=True, exist_ok=True)
        torch.save(data, path)
        return LoadMechanism.from_url_string(str(url), category)


class CacheObj(CacheCategory):
    @staticmethod
    def store(data: ELEMENT_DATA_TYPE, url: URIComponents | None, category: DataCategory) -> LoadMechanism:
        if url is None:
            return LoadMechanism(data, category)

        if url.scheme not in ["", "file"]:
            raise NotImplementedError()
        import pickle

        path = Path(str(url))

        Path.mkdir(path.parent, parents=True, exist_ok=True)
        with open(path, "wb") as f:
            pickle.dump(data, f)
        return LoadMechanism.from_url_string(str(url), category)


class CachingMethodExecutor:
    CACHE_DATA_METHODS: Dict[DataCategory, CacheCategory] = {
        DataCategory.image: CacheImage,
        DataCategory.torch: CacheTorch,
        DataCategory.obj: CacheObj,
    }

    def store(self, data: ELEMENT_DATA_TYPE, url: URIComponents | None, category: DataCategory) -> LoadMechanism:
        return self.CACHE_DATA_METHODS[category].store(data, url, category)



from pathlib import Path

import numpy as np
import pytest

from bridge.primitives.dataset import Dataset
from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
from bridge.primitives.element.data.category import DataCategory
from bridge.primitives.element.data.load.load_mechanism import LoadMechanism
from bridge.primitives.element.data.uri_components import URIComponents
from bridge.primitives.element.element import Element
from bridge.primitives.element.element_type import ElementType
from bridge.utils.constants import ELEMENT_COLS
from bridge.utils.data_objects import ClassLabel


@pytest.fixture
def dummy_elements():
    """
    sample ids: 0-100
    element ids: 0-100
    """
    elements = []
    for i in range(100):
        img_element = Element(
            element_id=i,
            sample_id=i,
            etype=ElementType.image,
            load_mechanism=LoadMechanism(
                url_or_data=np.random.randint(0, 255, size=(100, 100, 3)).astype("uint8"),
                category=DataCategory.obj,
            ),
        )
        lbl_element = Element(
            element_id=f"label_{i}",
            sample_id=i,
            etype=ElementType.class_label,
            load_mechanism=LoadMechanism(
                url_or_data=ClassLabel(class_idx=np.random.randint(0, 10)), category=DataCategory.obj
            ),
        )
        elements.extend([img_element, lbl_element])
    return elements


@pytest.fixture
def dummy_elements_2():
    """
    sample ids: 50-150 (50 sample id overlap)
    element ids: 100-200 (0 element id overlap)
    """
    elements = []
    for i in range(100):
        img_element = Element(
            element_id=100 + i,
            sample_id=50 + i,
            etype=ElementType.image,
            load_mechanism=LoadMechanism(
                url_or_data=np.random.randint(0, 255, size=(100, 100, 3)).astype("uint8"),
                category=DataCategory.obj,
            ),
        )
        lbl_element = Element(
            element_id=f"label_{100+i}",
            sample_id=50 + i,  # Adjusting sample_id to create overlap
            etype=ElementType.class_label,
            load_mechanism=LoadMechanism(
                url_or_data=ClassLabel(class_idx=np.random.randint(0, 10)), category=DataCategory.obj
            ),
        )
        elements.extend([img_element, lbl_element])
    return elements


@pytest.fixture
def dummy_dataset(dummy_elements):
    return Dataset.from_elements(dummy_elements)


@pytest.fixture
def dummy_dataset_2(dummy_elements_2):
    return Dataset.from_elements(dummy_elements_2)


@pytest.fixture
def dummy_classification_dataset_local_cache(tmp_path, dummy_elements):
    cache = CacheMechanism(URIComponents.from_str(str(tmp_path)))
    return Dataset.from_elements(elements=dummy_elements, cache_mechanisms={ElementType.image: cache})


def test_repr(dummy_dataset):
    lens_dict = {"n_samples": 100, "n_class_label": 100, "n_image": 100}
    assert "Dataset: " + str(lens_dict) == repr(dummy_dataset)


def test_select_by_etype(dummy_dataset):
    ds = dummy_dataset.select(lambda e: e[ELEMENT_COLS.ETYPE] == ElementType.class_label)
    assert len(ds) == 100
    assert ElementType.image not in ds.elements[ELEMENT_COLS.ETYPE].drop_duplicates().to_list()


def test_select_by_sample_id(dummy_dataset):
    sample_ids = np.random.choice(
        dummy_dataset.elements.index.get_level_values(ELEMENT_COLS.SAMPLE_ID).drop_duplicates(),
        len(dummy_dataset) // 2,
        replace=False,
    )

    ds = dummy_dataset.select(lambda e: sample_ids)
    assert len(ds) == len(dummy_dataset) // 2


def test_data(dummy_dataset):
    images = []
    labels = []
    for sample in dummy_dataset:
        data_dict = sample.data
        images.extend(data_dict[ElementType.image])
        labels.extend(data_dict[ElementType.class_label])
    assert all([isinstance(img, np.ndarray) for img in images])
    assert all([img.shape == (100, 100, 3) for img in images])
    assert all([isinstance(label, ClassLabel) for label in labels])


def test_cache(dummy_classification_dataset_local_cache):
    cache_path = str(dummy_classification_dataset_local_cache._cache_mechanisms[ElementType.image]._root_uri)

    # assert starting conditions: cache dir exists and is empty, all elements have object data rather than URI
    assert Path(cache_path).exists() and Path(cache_path).is_dir()
    assert len(list(Path(cache_path).iterdir())) == 0
    assert dummy_classification_dataset_local_cache.elements.data.apply(
        lambda d: not isinstance(d, URIComponents)
    ).all()

    # `sample.data` will call the cache mechanism for every element in the dataset, if exists
    for sample in dummy_classification_dataset_local_cache:
        _ = sample.data

    image_schemes = (
        dummy_classification_dataset_local_cache.elements.pipe(
            lambda df_: df_.loc[df_[ELEMENT_COLS.ETYPE] == ElementType.image]
        )
        .data.apply(lambda d: d.scheme)
        .drop_duplicates()
        .values
    )
    assert len(list(Path(cache_path).iterdir())) == len(dummy_classification_dataset_local_cache)
    assert len(image_schemes) == 1 and image_schemes[0] == ""


def test_merge_duplicate_elements(dummy_dataset):
    with pytest.raises(AssertionError):
        _ = dummy_dataset.merge(dummy_dataset)


def test_merge_datasets(dummy_dataset, dummy_dataset_2):
    merged_ds = dummy_dataset.merge(dummy_dataset_2)
    assert len(merged_ds) == 150
    assert len(merged_ds.elements) == 400



# Col Names
import dataclasses
from dataclasses import dataclass
from typing import List


def recursively_list_columns(col_names) -> List[str]:
    cols = []
    for col in dataclasses.fields(col_names):
        field = getattr(col_names, col.name)
        if col.type == str:
            cols.append(field)
        else:
            cols.extend(recursively_list_columns(field))
    return list(set(cols))


class FlatAttributeListMixin:
    def list(self) -> List[str]:
        return recursively_list_columns(self)


@dataclass
class LoadMechanismCols(FlatAttributeListMixin):
    CATEGORY: str
    URL_OR_DATA: str


@dataclass
class ElementCols(FlatAttributeListMixin):
    ID: str
    SAMPLE_ID: str
    ETYPE: str
    LOAD_MECHANISM: LoadMechanismCols


ELEMENT_ID_COL_NAME = "element_id"
SAMPLE_ID_COL_NAME = "sample_id"
DATA_CATEGORY_COL_NAME = "category"
ELEMENT_TYPE_COL_NAME = "element_type"

ELEMENT_DATA_COL_NAME = "data"


LOAD_MECHANISM_COL_NAMES = LoadMechanismCols(CATEGORY=DATA_CATEGORY_COL_NAME, URL_OR_DATA=ELEMENT_DATA_COL_NAME)

ELEMENT_COLS = ElementCols(
    ID=ELEMENT_ID_COL_NAME,
    SAMPLE_ID=SAMPLE_ID_COL_NAME,
    ETYPE=ELEMENT_TYPE_COL_NAME,
    LOAD_MECHANISM=LOAD_MECHANISM_COL_NAMES,
)


INDICES = [ELEMENT_COLS.SAMPLE_ID, ELEMENT_COLS.ID]

IS_SAMPLE_COL_NAME = "is_example"

__all__ = ["ELEMENT_COLS", "INDICES", "IS_SAMPLE_COL_NAME"]



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Callable, Dict, Hashable, Iterable, Iterator, Sequence

from tqdm.contrib import tmap
from typing_extensions import Self

if TYPE_CHECKING:
    from bridge.display.display_engine import DisplayEngine
    from bridge.primitives.element.data.cache.cache_mechanism import CacheMechanism
    from bridge.primitives.element.element import Element
    from bridge.primitives.element.element_type import ElementType
    from bridge.primitives.sample import Sample
    from bridge.primitives.sample.transform import SampleTransform


class SampleAPI(abc.ABC):
    @abc.abstractmethod
    def iget(self, index: int) -> Sample:
        pass

    @abc.abstractmethod
    def get(self, sample_id: Hashable) -> Sample:
        pass

    @abc.abstractmethod
    def transform_samples(
        self,
        transform: SampleTransform,
        map_fn=tmap,
        cache_mechanisms: Dict[ElementType, CacheMechanism] | None = None,
        display_engine: DisplayEngine | None = None,
    ) -> Self:
        pass

    @abc.abstractmethod
    def map_samples(self, function: Callable[[Sample], Any], map_fn=tmap) -> Sequence[Sample]:
        pass

    @abc.abstractmethod
    def __iter__(self) -> Iterator[Sample]:
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    @classmethod
    @abc.abstractmethod
    def from_elements(cls, elements: Iterable[Element]) -> Self:
        pass



from pathlib import Path

import pytest

NOTEBOOK_DIR = Path.cwd() / "docs" / "source" / "notebooks"
DATA_DIR = Path.cwd() / "tests" / "data"

NOTEBOOKS_LIST = [p for p in NOTEBOOK_DIR.rglob("*[!\.ipynb_checkpoints]*.ipynb") if ".ipynb_checkpoints" not in str(p)]


@pytest.fixture(
    params=NOTEBOOKS_LIST,
    ids=[p.name for p in NOTEBOOKS_LIST],
)
def tb(request):
    from testbook import testbook

    with testbook(request.param, timeout=-1) as tb:
        yield tb


# @pytest.mark.skip(reason="This test needs to be moved to the correct location")
def test_notebook(tb):
    with tb.patch("tempfile.mkdtemp", return_value=str(DATA_DIR)):
        for cell in tb.cells:
            if "!pip install bridge-ds" in cell["source"]:
                cell["source"] = cell["source"].replace("!pip install bridge-ds", "")
        tb.execute()



from __future__ import annotations

from bridge.utils import StrEnum


class ElementType(StrEnum):
    image = "image"
    class_label = "class_label"
    bbox = "bbox"
    text = "text"
    segmentation = "segmentation"
    keypoint = "keypoint"
